"""
fine-tuning playground
"""

# %%
import os
import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.linalg import norm

from libs.pipeline import *
from libs.visualizations import *
from fno.sfno import SFNO, SpectralConvS, OutConv
from fno.finetune import OutConvFT
from torch.utils.data import DataLoader
from libs.visualizations import *

get_seed(42)
# torch.cuda.manual_seed_all(0)

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
print(device)
for p in [SRC_ROOT, MODEL_PATH, DATA_PATH]:
    print(p)

# %%
VAL_PATH = TEST_PATH = os.path.join(
    DATA_PATH, "McWilliams2d_fp32_64x64_N1152_v1e-3_T100.pt"
)
TEST_PATH = os.path.join(
    DATA_PATH, "McWilliams2d_fp64_256x256_N16_v1e-3_T100.pt"
)
Ntrain = 1024
Ntest = 16
Nval = 64


batch_size = 1

epochs = 15
learning_rate = 0.01

n = 256
nc = 64
T = 40
fields = ["vorticity", "stream", "residual", "vort_t"]
fs = 'vorticity'
beta = -1e-2

modes = 32
modes_t = 5
width = 10
dtype = torch.float64

beta_str = f"{beta:.0e}".replace("e-0", "e-").replace("e+0", "e")
model_name = f"sfno_ex_McWilliams2d_ep{epochs}_m{modes}_w{width}_b{beta_str}.pt"
path_model = os.path.join(MODEL_PATH, model_name)
print(model_name)
torch.set_default_dtype(dtype)

# %% 

test_dataset = BochnerDataset(datapath=TEST_PATH, 
                              fields=fields,
                              n_samples=Ntest,
                              steps=T,
                              T_start=50,
                              dtype=dtype,
                              train=False)
test_loader = DataLoader(test_dataset, batch_size=1, shuffle=False)


# %%
torch.cuda.empty_cache()
model = SFNO(modes, modes, modes_t, width, beta).to(device)
print(get_num_params(model))

model.load_state_dict(torch.load(path_model))
#%%
for name, param in model.q.conv.named_parameters():
    print(f"{name:<15} | {','.join(str(i) for i in list(param.shape)):<15}| {param.dtype}")

#%%
idx = 2
model.eval()
model.add_latent_hook("r")

sample = test_dataset.__getitem__(idx = idx, start_steps=50)
vort_inp_gt = sample[0]['vorticity']
vort_inp_gt = vort_inp_gt.unsqueeze(0).to(device)

#%%

with timer("Inference", compact=True):
    with torch.no_grad():
        preds = model(vort_inp_gt, out_steps=40)
        preds_no = preds.detach()


v_latent = model.latent_tensors["r"]
print(v_latent.shape)
#%%

torch.cuda.empty_cache()
model = SFNO(modes, modes, modes_t, width, beta).to(device)
model.load_state_dict(torch.load(path_model))
model.to(dtype)
f = torch.zeros((n, n))[None, ...].to(device)

#%%

l2_loss_setting = {
    "n_grid": n,
    "relative": True,
    "time_average": True,
    "diam": 2 * torch.pi,
    "freq_cutoff": n // 2 + 1,
}

L2diff = SobolevLoss(**l2_loss_setting)

res_loss_setting = {
    "n_grid": n,
    "relative": False,
    "time_average": True,
    "diam": 2 * torch.pi,
    "freq_cutoff": n // 2 + 1,
    "alpha": 10**(-3/2),
}

ResHm1Loss = SobolevLoss(norm_order=-1, **res_loss_setting)
ResL2Loss = SobolevLoss(norm_order=0, **res_loss_setting)

loss_func = ResHm1Loss

modes_orig = (modes, modes, modes_t)
modes_ft = (64, 64, 6)
ft_kws = {"n_grid": n, 
          "dt": 1e-6,
          "delta": 1,
          "diam": 2*torch.pi,
          "finetune": True,
          "bdf_weight": (0.5, 0.5),
          }
qft = OutConvFT(*modes_orig,
                **ft_kws)
qft._update_spectral_conv_weights(*modes_ft, device=device, model=model.q)
qft.to(device).to(dtype)
optimizer = torch.optim.Adam([
            {'params': qft.conv.bias, 'lr': 1e-2},
            {'params': qft.conv.weight, 'lr': 1e-5},
        ])

preds_ft_history = []
l2diff_ft_gt = []
residual_hm1norms = []
residual_l2norms = []
qft_bias_norms = []
qft_weight_norms = []



#%%
num_iter = 50

with tqdm(total=num_iter) as pbar:
    for i in range(num_iter):
        qft.train()

        def closure():
            optimizer.zero_grad()
            preds_ft_all = qft(v_latent, vort_inp_gt, f, out_steps=T)
            res_ft_new = preds_ft_all["residual"]
            loss = loss_func(res_ft_new)
            loss.backward(retain_graph=True)
            return loss

        preds_ft_all = qft(v_latent, vort_inp_gt, f, out_steps=T)
        preds_ft = preds_ft_all["w"]
        res_ft = preds_ft_all["residual"]

        reshm1loss = ResHm1Loss(res_ft)
        resl2loss = ResL2Loss(res_ft)

        optimizer.step(closure)


        optimizer.zero_grad()

        preds_ft_history.append(preds_ft.clone().detach().cpu())

        residual_hm1norms.append(reshm1loss.item())
        residual_l2norms.append(resl2loss.item())


        pbar.set_description(
            f"Res Hm1 : {reshm1loss.item():.3e} | Res L2 : {resl2loss.item():.3e}"
        )
        pbar.update()
